/***************************************************************************
 *   Copyright (C) 2008  Antony Martin <martin(dot)antony(at)yahoo(dot)fr> *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or         *
 *   modify it under the terms of the GNU General Public License           *
 *   as published by the Free Software Foundation; either version 2        *
 *   of the License, or (at your option) any later version.                *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA          *
 ***************************************************************************/

/* Cree le : 19 janvier 2007
   derniere modification le 12/07/2008 */

#ifndef SCEMESH_H
#define SCEMESH_H

#include <SCE/utils/SCEVector.h>
#include <SCE/utils/SCEList.h>
#include <SCE/core/SCECBuffers.h>
#include <SCE/interface/SCEBoundingBox.h>

#ifdef __cplusplus
extern "C"
{
#endif

/* types de vertices pouvant etre generes (AddGenVertices()) */
#define SCE_GEN_TANGENTS 0x00000001
#define SCE_GEN_BINORMALS 0x00000002
#define SCE_GEN_NORMALS 0x00000004

/* types de tri */
#define SCE_MESH_SORT_NEAR_TO_FAR 1
#define SCE_MESH_SORT_FAR_TO_NEAR 2


/* type des vecteurs des sommets */
typedef GLfloat SCEvertices;
#define SCE_VERTICES_TYPE SCE_FLOAT

/* type des indices */
typedef GLuint SCEindices;
#define SCE_INDICES_TYPE SCE_UNSIGNED_INT

/**
 * \copydoc sce_smeshvertexdata
 */
typedef struct sce_smeshvertexdata SCE_SMeshVertexData;
/**
 * \brief Structure that stores vertices' data
 * \sa SCE_CVertexDeclaration
 */
struct sce_smeshvertexdata
{
    SCE_CVertexDeclaration *dec; /**< Vertex declaration */
    void *data;                  /**< Data */
    int canfree;                 /**< Do we have right to delete \c data ? */
    size_t size;                 /**< Size of \c data in bytes */
};

/**
 * \copydoc sce_smeshvertexbuffer
 */
typedef struct sce_smeshvertexbuffer SCE_SMeshVertexBuffer;
/**
 * \brief Vertex buffer that stores many informations of the data's arrengment
 * \sa SCE_SMeshVertexData SCE_CVertexBuffer
 */
struct sce_smeshvertexbuffer
{
    SCEenum usage;              /**< Buffer's usage (GL_*_DRAW) */
    SCE_CVertexBuffer *buffer;  /**< The vertex buffer */
    SCE_SList *data;            /**< List of the \c buffer data, contains
                                 * SCE_SMeshVertexData structures */
    int active;                 /**< Is buffer actived ? */
    int builded;                /**< Is buffer builded ? */
};

/**
 * \copydoc sce_smeshindexbuffer
 */
typedef struct sce_smeshindexbuffer SCE_SMeshIndexBuffer;
/**
 * \brief Index buffer that stores many informations of the data's arrengment
 * \sa SCE_CIndexBuffer
 */
struct sce_smeshindexbuffer
{
    SCEenum usage;            /**< Buffer's usage (GL_*_DRAW) */
    SCE_CIndexBuffer *buffer; /**< The index buffer */
    SCEenum type;             /**< Data type (GL_UNSIGNED_INT, ...) */
    void *data;               /**< Data */
    int canfree;              /**< Do we have right to delete \c data ? */
    size_t size;              /**< Size of \c data in bytes */
    int builded;              /**< Is buffer builded ? */
};


/**
 * \copydoc sce_smesh
 */
typedef struct sce_smesh SCE_SMesh;
/**
 * \brief A mesh structure that contains all geometry informations about a mesh
 * \sa SCE_SMeshVertexBuffer SCE_SMeshIndexBuffer
 */
struct sce_smesh
{
    SCE_SList *vertices;     /**< List that contains all the vertex buffers of
                              * the mesh (elements are SCE_SMeshVertexBuffer) */
    SCE_SMeshIndexBuffer ib; /**< Index buffer */
    int use_indices;         /**< Do we use indices for rendering ? */

    unsigned int vcount;     /**< Number of vertices */
    unsigned int icount;     /**< Number of indices */

    SCEenum polygon_type;    /**< Polygons' type (triangle, quad, ...) */

    int builded;             /**< Is mesh builded ? */

    SCEvertices *pos,        /**< Pointer to the vertices' positions */
                *nor;        /**< Pointer to the vertices' normals */

    SCE_SList *sortedfaces;  /**< Sorted list that stores all the faces from
                              * his indices */
};

/**
 * \copydoc sce_smeshface
 */
typedef struct sce_smeshface SCE_SMeshFace;
/**
 * \brief A structure of a face
 */
struct sce_smeshface
{
    SCEindices index[4]; /**< Indices of the face, to get the n-th vertex's
                          * position of the face: pos[index[n]] */
    SCEvertices *pos,    /**< Pointer to the vertices' positions */
                *nor;    /**< Pointer to the vertices' normals */
    int num_vertices;    /**< Number of vertices of the mesh */
};

/**
 * \brief Function which is called for each face of a mesh
 */
typedef int (*SCE_FMeshFaceFunc)(SCE_SMesh*, SCE_SMeshFace*, int, void*);


/* initialise le gestionnaire de meshs */
int SCE_Init_Mesh (void);
/* quitte le gestionnaire */
void SCE_Quit_Mesh (void);

void SCE_Mesh_InitVB (SCE_SMeshVertexBuffer*);
void SCE_Mesh_InitIB (SCE_SMeshIndexBuffer*);

SCE_SMeshVertexBuffer* SCE_Mesh_CreateVB (void);
void SCE_Mesh_DeleteVB (void*);

int SCE_Mesh_BuildVB (SCE_SMeshVertexBuffer*);
int SCE_Mesh_BuildIB (SCE_SMeshIndexBuffer*);

/* cree et initialise un nouveau mesh */
SCE_SMesh* SCE_Mesh_Create (void);
/* supprime un mesh existant */
void SCE_Mesh_Delete (SCE_SMesh*);
/* supprime un tableau de meshs */
void SCE_Mesh_DeleteList (SCE_SMesh**);

/* ajoute un vertex buffer */
int SCE_Mesh_AddVB (SCE_SMesh*, SCE_SMeshVertexBuffer*);
/* ajoute un nouveau vertex buffer */
int SCE_Mesh_AddNewVB (SCE_SMesh*, int);
/* enleve un vertex buffer */
SCE_SMeshVertexBuffer* SCE_Mesh_RemoveVB (SCE_SMesh*, unsigned int);
/* supprime un vertex buffer */
void SCE_Mesh_EraseVB (SCE_SMesh*, unsigned int);

/* defini l'etat du tampon de sommets specifie */
void SCE_Mesh_ActivateVB (SCE_SMesh*, unsigned int, int);
/* defini l'etat du tampon d'indices */
void SCE_Mesh_ActivateIB (SCE_SMesh*, int);

/* recherche des donnees dans la liste de vertices (SCE_SList) */
SCE_SMeshVertexData* SCE_Mesh_LocateData (SCE_SMesh*, int,
                                          SCE_SMeshVertexBuffer**);
    /* retourne les positions de sommet du mesh */
void* SCE_Mesh_GetVerticesPositions (SCE_SMesh*);
void* SCE_Mesh_GetVerticesNormals (SCE_SMesh*);

/* envoi un tableau de donnees d'un type specifique au mesh specifie */
int SCE_Mesh_AddVertices (SCE_SMesh*, unsigned int, int, SCEenum,
                          unsigned int, unsigned int, void*, int);
/* idem, mais duplique les donnees et les convertis en SCEvertices */
int SCE_Mesh_AddVerticesDup (SCE_SMesh*, unsigned int, int, SCEenum,
                             unsigned int, unsigned int, void*);

/* attribue un tableau d'indices au mesh specifie */
int SCE_Mesh_SetIndices (SCE_SMesh*, SCEenum, SCEenum,
                         unsigned int, void*, int);
/* idem, mais duplique les donnees en les convertissant dans le meilleur type */
int SCE_Mesh_SetIndicesDup (SCE_SMesh*, SCEenum, SCEenum, unsigned int, void*);

/* retourne le nombre de vertices par face d'un mesh */
int SCE_Mesh_GetNumVerticesPerFace (SCE_SMesh*);
/* retourne le nombre de faces d'un mesh */
int SCE_Mesh_GetNumFaces (SCE_SMesh*);

/* effectue une operation pour chaque face d'un mesh */
int SCE_Mesh_ForEachFace (SCE_SMesh*, SCE_FMeshFaceFunc, void*);

/* defini l'ordre dans lequel seront rendus les polygones */
int SCE_Mesh_SortFaces (SCE_SMesh*, int, SCE_TVector3);

/* fonction de calcul des tangentes et des binormales pour un triangle donne.
   vertex et texcoord sont des tableaux dans lesquels la fonction se deplace
   librement selon les indices stockes dans index.
   la fonction ne depasse jamais index[2] */
void SCE_Mesh_ComputeTriangleTBN (
  SCEvertices*, SCEvertices*, SCEindices*,   /* vertex, texcoord, index */
  SCEvertices*, SCEvertices*, SCEvertices*); /* tangents, binormals, normals */

/* fonction qui genere les tangentes, binormales et normales pour des vertices
   et coordonnees de texture donnes. un tableau d'indices peut etre fourni
   en option, s'il n'y en a pas, les sommets seront pris l'un apres l'autre.
   la fonction a egalement besoin de connaitre le nombre de sommet que contient
   vertex et texcoord, et le nombre d'indices disponibles dans indices.
   le type des polygones est egalement requis, si celui-ci n'est pas supporte,
   une erreur est generee et la fonction renvoie SCE_ERROR. */
int SCE_Mesh_ComputeTBN (
  SCEenum,
  SCEvertices*, SCEvertices*, SCEindices*,   /* vertex, texcoord, index */
  unsigned int, unsigned int,                /* icount, vcount */
  SCEvertices*, SCEvertices*, SCEvertices*); /* tangents, binormals, normals */

/* genere les vecteurs tangents, binormaux et normaux d'un mesh donne.
   l'avantage a utiliser cette fonction est qu'elle prend en compte le type
   des variables des tableaux de sommets du mesh. */
int SCE_Mesh_GenerateTBN (SCE_SMesh*, SCEvertices*, SCEvertices*, SCEvertices*,
                          unsigned int);

/* equivalent a GenerateTBN, mais ajoute directement les donnees au mesh */
int SCE_Mesh_AddGenVertices (
  SCE_SMesh*,
  SCEuint,                /* id du tampon de stockage dans le mesh */
  SCEenum,                /* type des vertices generes (t, b, n...) */
  SCEuint,                /* type des donnees (attrib, ...) */
  SCEuint);               /* unite de texture */


/* calcul le plus petit parrallelepipede rectangle qui englobe le mesh */
void SCE_Mesh_ComputeBoundingBox (SCEvertices*, unsigned int,
                                  SCE_TVector3, float*, float*, float*);
void SCE_Mesh_ComputeBoundingBoxv (SCEvertices*, unsigned int,
                                   SCE_TVector3, SCE_TVector3);
int SCE_Mesh_GenerateBoundingBox (SCE_SMesh*, SCE_SBoundingBox*);
/* calcul la plus petite sphere qui englobe le mesh */
void SCE_Mesh_ComputeBoundingSphere (SCEvertices*, unsigned int,
                                     SCE_TVector3, float*);
int SCE_Mesh_GenerateBoundingSphere (SCE_SMesh*, SCE_TVector3, float*);

/* cree un parrallelepipede rectangle a partir
   d'un point d'origine et de ses dimensions */
SCE_SMesh* SCE_Mesh_CreateCube (SCE_TVector3, float, float, float);
SCE_SMesh* SCE_Mesh_CreateCubev (SCE_TVector3, SCE_TVector3);
/* version utilisant les indices */
SCE_SMesh* SCE_Mesh_CreateIndexedCube (SCE_TVector3, float, float, float);
SCE_SMesh* SCE_Mesh_CreateIndexedCubev (SCE_TVector3, SCE_TVector3);
#if 0
/* cree un mesh a partir de la bounding box d'un autre mesh */
SCE_SMesh* SCE_Mesh_CreateFromBoundingBox (SCE_SBoundingBox*);
#endif

/* construit le mesh specifie. il est possible de reappeler
   Build si l'on a ajoute/enleve des donnees a un ou plusieurs buffers */
int SCE_Mesh_Build (SCE_SMesh*);

/* defini le type mode de rendu du mesh (type de polygone) */
void SCE_Mesh_SetRenderMode (SCE_SMesh*, SCEenum);

/* met a jour les donnees du tampon specifie */
int SCE_Mesh_Update (SCE_SMesh*, unsigned int);
/* enregistre une mise a jour a effectuer ulterieurement */
int SCE_Mesh_LogUpdate (SCE_SMesh*, unsigned int);
/* lance toutes les mises a jour enregistrees */
void SCE_Mesh_ApplyUpdates (void);

/* charge tous les meshs contenus dans un fichier */
SCE_SMesh** SCE_Mesh_Load (const char*, int*);

/* stock et defini le mesh specifie comme mesh actif pour le rendu */
void SCE_Mesh_Use (SCE_SMesh*);
/* effectue le rendu du mesh specifie par Use */
void SCE_Mesh_Draw (void);
/* rendu d'une plage d'un tableau de meshs.
   la plage est definie tel qu'elle commence a first et fini a nb-1 */
void SCE_Mesh_Render (SCE_SMesh*);
/* rend tous les meshs compris dans l'invervalle 'start' 'n' dans 'meshs' */
void SCE_Mesh_RenderList (SCE_SMesh**, unsigned int, unsigned int);

#ifdef __cplusplus
} /* extern "C" */
#endif

#endif /* guard */
